#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import datetime
import warnings
from typing import Any, Dict, List, Optional, TypeVar, Union

import pandas as pd
from hopsworks_common import client
from hopsworks_common.client.exceptions import FeatureStoreException
from hopsworks_common.core.constants import HAS_NUMPY
from hsfs import (
    engine,
    feature_group,
    feature_view,
    training_dataset_feature,
    util,
)
from hsfs.client import exceptions
from hsfs.constructor.filter import Filter, Logic
from hsfs.core import (
    feature_view_api,
    query_constructor_api,
    statistics_engine,
    tags_api,
    training_dataset_engine,
    transformation_function_engine,
)
from hsfs.core.feature_logging import FeatureLogging
from hsfs.feature_logger import FeatureLogger
from hsfs.training_dataset_split import TrainingDatasetSplit


if HAS_NUMPY:
    import numpy as np


class FeatureViewEngine:
    ENTITY_TYPE = "featureview"
    _TRAINING_DATA_API_PATH = "trainingdatasets"
    _OVERWRITE = "overwrite"
    _APPEND = "append"

    _LOG_TD_VERSION = "td_version"
    _LOG_TIME = "log_time"
    _HSML_MODEL = "hsml_model"

    def __init__(self, feature_store_id):
        self._feature_store_id = feature_store_id

        self._feature_view_api = feature_view_api.FeatureViewApi(feature_store_id)
        self._tags_api = tags_api.TagsApi(feature_store_id, self.ENTITY_TYPE)
        self._statistics_engine = statistics_engine.StatisticsEngine(
            feature_store_id, self._TRAINING_DATA_API_PATH
        )
        self._training_dataset_engine = training_dataset_engine.TrainingDatasetEngine(
            feature_store_id
        )
        self._transformation_function_engine = (
            transformation_function_engine.TransformationFunctionEngine(
                feature_store_id
            )
        )
        self._query_constructor_api = query_constructor_api.QueryConstructorApi()

    def save(
        self, feature_view_obj: feature_view.FeatureView
    ) -> feature_view.FeatureView:
        """
        Save a feature view to the backend.

        # Arguments
            feature_view_obj `FeatureView` : The feature view object to be saved.

        # Returns
            `FeatureView` : Updated feature view that has the ID used to save in the backend.
        """
        if feature_view_obj.query.is_time_travel():
            warnings.warn(
                "`as_of` argument in the `Query` will be ignored because"
                " feature view does not support time travel query.",
                stacklevel=1,
            )
        if feature_view_obj.labels:
            for label_name in feature_view_obj.labels:
                (
                    feature,
                    prefix,
                    featuregroup,
                ) = feature_view_obj.query._get_feature_by_name(label_name)
                feature_view_obj._features.append(
                    training_dataset_feature.TrainingDatasetFeature(
                        name=feature.name,
                        label=True,
                        featuregroup=featuregroup,
                    )
                )
        if feature_view_obj.inference_helper_columns:
            for helper_column_name in feature_view_obj.inference_helper_columns:
                (
                    feature,
                    prefix,
                    featuregroup,
                ) = feature_view_obj.query._get_feature_by_name(helper_column_name)
                feature_view_obj._features.append(
                    training_dataset_feature.TrainingDatasetFeature(
                        name=feature.name,
                        inference_helper_column=True,
                        featuregroup=featuregroup,
                    )
                )

        if feature_view_obj.training_helper_columns:
            for helper_column_name in feature_view_obj.training_helper_columns:
                (
                    feature,
                    prefix,
                    featuregroup,
                ) = feature_view_obj.query._get_feature_by_name(helper_column_name)
                feature_view_obj._features.append(
                    training_dataset_feature.TrainingDatasetFeature(
                        name=feature.name,
                        training_helper_column=True,
                        featuregroup=featuregroup,
                    )
                )

        updated_fv = self._feature_view_api.post(feature_view_obj)
        print(
            "Feature view created successfully, explore it at \n"
            + self._get_feature_view_url(updated_fv)
        )
        return updated_fv

    def update(
        self, feature_view_obj: feature_view.FeatureView
    ) -> feature_view.FeatureView:
        """
        Update the feature view object saved in the backend

        # Arguments
            feature_view_obj `FeatureView` : The feature view object to be saved.

        # Returns
            `FeatureView` : Updated feature view that has the ID used to save in the backend.
        """
        self._feature_view_api.update(feature_view_obj)
        return feature_view_obj

    def get(
        self, name: str, version: int = None
    ) -> Union[feature_view.FeatureView, List[feature_view.FeatureView]]:
        """
        Get a feature view form the backend using name or using name and version.

        If version is not provided then a List of feature views containing all of its versions is returned.

        # Arguments
            name `str`: Name of feature view.
            version `version`: Version of the feature view.

        # Returns
            `Union[FeatureView, List[FeatureView]]`

        # Raises
            `RestAPIError`: If the feature view cannot be found from the backend.
            `ValueError`: If the feature group associated with the feature view cannot be found.
        """
        if version:
            fv = self._feature_view_api.get_by_name_version(name, version)
        else:
            fv = self._feature_view_api.get_by_name(name)
        return fv

    def delete(self, name, version=None):
        if version:
            return self._feature_view_api.delete_by_name_version(name, version)
        else:
            return self._feature_view_api.delete_by_name(name)

    def get_training_dataset_schema(
        self,
        feature_view: feature_view.FeatureView,
        training_dataset_version: Optional[int] = None,
    ):
        """
        Function that returns the schema of the training dataset generated using the feature view.

        # Arguments
            feature_view: `FeatureView`. The feature view for which the schema is to be generated.
            training_dataset_version: `int`. Specifies the version of the training dataset for which the schema should be generated.
                By default, this is set to None. However, if the `one_hot_encoder` transformation function is used, the training dataset version must be provided.
                This is because the schema will then depend on the statistics of the training data used.

        # Returns
            `List[training_dataset_feature.TrainingDatasetFeature]`: List of training dataset features objects.

        # Raises
            `ValueError` if the  training dataset version provided cannot be found.
        """

        if not feature_view.transformation_functions:
            # Return the features in the feature view if the no transformation functions in the feature view.
            return feature_view.features
        else:
            transformed_features = []
            transformed_labels = []
            dropped_features = set()

            # Statistics only required for computing schema if one-hot-encoder in the transformation functions
            statistics_required = any(
                [
                    tf.hopsworks_udf.function_name == "one_hot_encoder"
                    for tf in feature_view.transformation_functions
                ]
            )

            if statistics_required:
                if not training_dataset_version:
                    raise FeatureStoreException(
                        "The feature view includes the one_hot_encoder transformation function. As a result, the schema of the generated training dataset depends on its statistics. Please specify the version of the training dataset for which the schema should be generated."
                    )

                # Get transformation functions with correct statistics based on training dataset version.
                transformation_functions = self._transformation_function_engine.get_ready_to_use_transformation_fns(
                    feature_view=feature_view,
                    training_dataset_version=training_dataset_version,
                )
            else:
                transformation_functions = feature_view.transformation_functions

            # Getting all dropped features
            for tf in transformation_functions:
                if tf.hopsworks_udf.dropped_features:
                    dropped_features.update(tf.hopsworks_udf.dropped_features)

            # Creating list of features not dropped after transformations
            transformed_features = [
                feature
                for feature in feature_view.features
                if feature.name not in dropped_features and not feature.label
            ]

            # Creating list of labels not dropped after transformations
            transformed_labels = [
                feature
                for feature in feature_view.features
                if feature.name not in dropped_features and feature.label
            ]

            for tf in transformation_functions:
                # create transformed labels if a transformation functions take a labels as input.
                if any(
                    label in tf.hopsworks_udf.transformation_features
                    for label in feature_view.labels
                ):
                    transformed_labels.extend(
                        [
                            training_dataset_feature.TrainingDatasetFeature(
                                name=transformed_label_name,
                                type=output_type,
                                label=True,
                            )
                            for transformed_label_name, output_type in zip(
                                tf.output_column_names, tf.hopsworks_udf.return_types
                            )
                        ]
                    )
                # create transformed features if a transformation functions take a no labels as input.
                else:
                    transformed_features.extend(
                        [
                            training_dataset_feature.TrainingDatasetFeature(
                                name=transformed_feature_name,
                                type=output_type,
                                label=False,
                            )
                            for transformed_feature_name, output_type in zip(
                                tf.output_column_names, tf.hopsworks_udf.return_types
                            )
                        ]
                    )

            return transformed_features + transformed_labels

    def get_batch_query(
        self,
        feature_view_obj,
        start_time,
        end_time,
        with_label=False,
        primary_keys=False,
        event_time=False,
        inference_helper_columns=False,
        training_helper_columns=False,
        training_dataset_version=None,
        spine=None,
    ):
        try:
            query = self._feature_view_api.get_batch_query(
                feature_view_obj.name,
                feature_view_obj.version,
                util.convert_event_time_to_timestamp(start_time),
                util.convert_event_time_to_timestamp(end_time),
                training_dataset_version=training_dataset_version,
                is_python_engine=engine.get_type() == "python",
                with_label=with_label,
                primary_keys=primary_keys,
                event_time=event_time,
                inference_helper_columns=inference_helper_columns,
                training_helper_columns=training_helper_columns,
            )
            # verify whatever is passed 1. spine group with dataframe contained, or 2. dataframe
            # the schema has to be consistent

            # allow passing new spine group or dataframe
            if isinstance(spine, feature_group.SpineGroup):
                # schema of original fg on left side needs to be consistent with schema contained in the
                # spine group to overwrite the feature group
                dataframe_features = engine.get_instance().parse_schema_feature_group(
                    spine.dataframe
                )
                spine._feature_group_engine._verify_schema_compatibility(
                    query._left_feature_group.features, dataframe_features
                )
                query._left_feature_group = spine
            elif isinstance(query._left_feature_group, feature_group.SpineGroup):
                if spine is None:
                    raise FeatureStoreException(
                        "Feature View was created with a spine group, setting the `spine` argument is mandatory."
                    )
                # the dataframe setter will verify the schema of the dataframe
                query._left_feature_group.dataframe = spine
            return query
        except exceptions.RestAPIError as e:
            if e.response.json().get("errorCode", "") == 270172:
                raise ValueError(
                    "Cannot generate dataset(s) from the given start/end time because"
                    " event time column is not available in the left feature groups."
                    " A start/end time should not be provided as parameters."
                ) from e
            else:
                raise e

    def get_batch_query_string(
        self, feature_view_obj, start_time, end_time, training_dataset_version=None
    ):
        try:
            query_obj = self._feature_view_api.get_batch_query(
                feature_view_obj.name,
                feature_view_obj.version,
                util.convert_event_time_to_timestamp(start_time),
                util.convert_event_time_to_timestamp(end_time),
                training_dataset_version=training_dataset_version,
                is_python_engine=engine.get_type() == "python",
            )
        except exceptions.RestAPIError as e:
            if e.response.json().get("errorCode", "") == 270172:
                raise ValueError(
                    "Cannot generate a query from the given start/end time because"
                    " event time column is not available in the left feature groups."
                    " A start/end time should not be provided as parameters."
                ) from e
            else:
                raise e

        fs_query = self._query_constructor_api.construct_query(query_obj)
        if fs_query.pit_query is not None:
            return fs_query.pit_query
        return fs_query.query

    def create_training_dataset(
        self,
        feature_view_obj,
        training_dataset_obj,
        user_write_options,
        spine=None,
        primary_keys=False,
        event_time=False,
        training_helper_columns=False,
    ):
        self._set_event_time(feature_view_obj, training_dataset_obj)
        updated_instance = self._create_training_data_metadata(
            feature_view_obj, training_dataset_obj
        )
        td_job = self.compute_training_dataset(
            feature_view_obj,
            user_write_options,
            training_dataset_obj=training_dataset_obj,
            spine=spine,
            primary_keys=primary_keys,
            event_time=event_time,
            training_helper_columns=training_helper_columns,
        )
        return updated_instance, td_job

    def get_training_data(
        self,
        feature_view_obj: feature_view.FeatureView,
        read_options=None,
        splits=None,
        training_dataset_obj=None,
        training_dataset_version=None,
        spine=None,
        primary_keys=False,
        event_time=False,
        training_helper_columns=False,
        dataframe_type="default",
    ):
        # check if provided td version has already existed.
        if training_dataset_version:
            td_updated = self._get_training_dataset_metadata(
                feature_view_obj, training_dataset_version
            )
        else:
            self._set_event_time(feature_view_obj, training_dataset_obj)
            td_updated = self._create_training_data_metadata(
                feature_view_obj, training_dataset_obj
            )
        # check splits
        if splits is None:
            splits = []
        if len(splits) != len(td_updated.splits):
            if len(td_updated.splits) == 0:
                method_name = "get_training_data"
            elif len(td_updated.splits) == 2:
                method_name = "get_train_test_split"
            elif len(td_updated.splits) == 3:
                method_name = "get_train_validation_test_split"
            raise ValueError(
                f"Incorrect `get` method is used. Use `feature_view.{method_name}` instead."
            )

        read_options = engine.get_instance().read_options(
            td_updated.data_format, read_options
        )

        if td_updated.training_dataset_type != td_updated.IN_MEMORY:
            split_df = self._read_from_storage_connector(
                td_updated,
                td_updated.splits,
                read_options,
                with_primary_keys=primary_keys,
                # at this stage training dataset was already written and if there was any name clash it should have
                # already failed in creation phase, so we don't need to check it here. This is to make
                # sure that both it is backwards compatible and also to drop columns if user set
                # primary_keys and event_times to True on creation stage.
                primary_keys=self._get_primary_keys_from_query(
                    feature_view_obj.query, False
                ),
                with_event_time=event_time,
                event_time=self._get_eventtimes_from_query(
                    feature_view_obj.query, False
                ),
                with_training_helper_columns=training_helper_columns,
                training_helper_columns=feature_view_obj.training_helper_columns,
                feature_view_features=[
                    feature.name for feature in feature_view_obj.features
                ],
                # forcing dataframe type to default here since dataframe operations are required for training data split.
                dataframe_type="default"
                if dataframe_type.lower() in ["numpy", "python"]
                else dataframe_type,  # forcing dataframe type to default here since dataframe operations are required for training data split.
            )
        else:
            self._check_feature_group_accessibility(feature_view_obj)
            query = self.get_batch_query(
                feature_view_obj,
                training_dataset_version=td_updated.version,
                start_time=td_updated.event_start_time,
                end_time=td_updated.event_end_time,
                with_label=True,
                inference_helper_columns=False,
                primary_keys=primary_keys,
                event_time=event_time,
                training_helper_columns=training_helper_columns,
                spine=spine,
            )
            split_df = engine.get_instance().get_training_data(
                td_updated,
                feature_view_obj,
                query,
                read_options,
                dataframe_type,
                training_dataset_version,
            )
            self.compute_training_dataset_statistics(
                feature_view_obj, td_updated, split_df
            )

        # Getting transformed label names
        transformed_labels = [
            feature.name
            for feature in self.get_training_dataset_schema(
                feature_view=feature_view_obj,
                training_dataset_version=td_updated.version,
            )
            if feature.label
        ]

        # Updating labels based if transformation functions are attached to the feature view.
        labels = (
            transformed_labels
            if feature_view_obj.transformation_functions
            else feature_view_obj.labels
        )

        # Set training dataset schema after training dataset has been generated
        td_updated.schema = self.get_training_dataset_schema(
            feature_view=feature_view_obj, training_dataset_version=td_updated.version
        )

        # split df into features and labels df
        if td_updated.splits:
            for split in td_updated.splits:
                split_name = split.name
                split_df[split_name] = engine.get_instance().split_labels(
                    split_df[split_name], labels, dataframe_type
                )
            feature_dfs = []
            label_dfs = []
            for split in splits:
                feature_dfs.append(split_df[split][0])
                label_dfs.append(split_df[split][1])
            return td_updated, feature_dfs + label_dfs
        else:
            split_df = engine.get_instance().split_labels(
                split_df, labels, dataframe_type
            )
            return td_updated, split_df

    def _set_event_time(self, feature_view_obj, training_dataset_obj):
        event_time = feature_view_obj.query._left_feature_group.event_time
        if event_time:
            if training_dataset_obj.splits:
                for split in training_dataset_obj.splits:
                    if (
                        split.split_type == TrainingDatasetSplit.TIME_SERIES_SPLIT
                        and split.name == TrainingDatasetSplit.TRAIN
                        and not split.start_time
                    ):
                        split.start_time = self._get_start_time()
                    if (
                        split.split_type == TrainingDatasetSplit.TIME_SERIES_SPLIT
                        and split.name == TrainingDatasetSplit.TEST
                        and not split.end_time
                    ):
                        split.end_time = self._get_end_time()
            else:
                if not training_dataset_obj.event_start_time:
                    training_dataset_obj.event_start_time = self._get_start_time()
                if not training_dataset_obj.event_end_time:
                    training_dataset_obj.event_end_time = self._get_end_time()

    def _get_start_time(self):
        # minimum start time is 1 second
        return 1000

    def _get_end_time(self):
        # end time is current time
        return int(float(datetime.datetime.now().timestamp()) * 1000)

    def recreate_training_dataset(
        self,
        feature_view_obj,
        training_dataset_version,
        statistics_config,
        user_write_options,
        spine=None,
    ):
        training_dataset_obj = self._get_training_dataset_metadata(
            feature_view_obj, training_dataset_version
        )

        if statistics_config is not None:
            # update statistics config if provided. This is currently the only way to update TD statistics config.
            # recreating a training dataset may result in different statistics if the FGs data in the FV query have been update.
            training_dataset_obj.statistics_config = statistics_config
            training_dataset_obj.update_statistics_config()

        td_job = self.compute_training_dataset(
            feature_view_obj,
            user_write_options,
            training_dataset_obj=training_dataset_obj,
            spine=spine,
        )
        # Set training dataset schema after training dataset has been generated
        training_dataset_obj.schema = self.get_training_dataset_schema(
            feature_view=feature_view_obj,
            training_dataset_version=training_dataset_obj.version,
        )
        return training_dataset_obj, td_job

    def _read_from_storage_connector(
        self,
        training_data_obj,
        splits,
        read_options,
        with_primary_keys,
        primary_keys,
        with_event_time,
        event_time,
        with_training_helper_columns,
        training_helper_columns,
        feature_view_features,
        dataframe_type,
    ):
        if splits:
            result = {}
            for split in splits:
                path = training_data_obj.location + "/" + str(split.name)
                result[split.name] = self._read_dir_from_storage_connector(
                    training_data_obj,
                    path,
                    read_options,
                    with_primary_keys,
                    primary_keys,
                    with_event_time,
                    event_time,
                    with_training_helper_columns,
                    training_helper_columns,
                    feature_view_features,
                    dataframe_type,
                )
            return result
        else:
            path = training_data_obj.location + "/" + training_data_obj.name
            return self._read_dir_from_storage_connector(
                training_data_obj,
                path,
                read_options,
                with_primary_keys,
                primary_keys,
                with_event_time,
                event_time,
                with_training_helper_columns,
                training_helper_columns,
                feature_view_features,
                dataframe_type,
            )

    def _cast_columns(self, data_format, df, schema):
        if data_format == "csv" or data_format == "tsv":
            if not schema:
                raise FeatureStoreException("Reading csv, tsv requires a schema.")
            return engine.get_instance().cast_columns(df, schema)
        else:
            return df

    def _read_dir_from_storage_connector(
        self,
        training_data_obj,
        path,
        read_options,
        with_primary_keys,
        primary_keys,
        with_event_time,
        event_time,
        with_training_helper_columns,
        training_helper_columns,
        feature_view_features,
        dataframe_type,
    ):
        try:
            df = training_data_obj.storage_connector.read(
                # always read from materialized dataset, not query object
                query=None,
                data_format=training_data_obj.data_format,
                options=read_options,
                path=path,
                dataframe_type=dataframe_type,
            )

            df = self._drop_helper_columns(
                df,
                feature_view_features,
                with_primary_keys,
                primary_keys,
                False,
                dataframe_type,
            )
            df = self._drop_helper_columns(
                df,
                feature_view_features,
                with_event_time,
                event_time,
                False,
                dataframe_type,
            )
            df = self._drop_helper_columns(
                df,
                feature_view_features,
                with_training_helper_columns,
                training_helper_columns,
                True,
                dataframe_type,
            )
            return df

        except Exception as e:
            if isinstance(e, FileNotFoundError):
                raise FileNotFoundError(
                    f"Failed to read dataset from {path}."
                    " Check if path exists or recreate a training dataset."
                ) from e
            else:
                raise e

    def _drop_helper_columns(
        self,
        df,
        feature_view_features,
        with_columns,
        columns,
        training_helper,
        dataframe_type,
    ):
        if not with_columns:
            if (
                engine.get_type().startswith("spark")
                and dataframe_type.lower() == "spark"
            ):
                existing_cols = [field.name for field in df.schema.fields]
            else:
                existing_cols = df.columns
            # primary keys and event time are dropped only if they are in the query
            drop_cols = list(set(existing_cols).intersection(columns))
            # training helper is always in the query
            if not training_helper:
                drop_cols = list(set(drop_cols).difference(feature_view_features))
            if drop_cols:
                df = engine.get_instance().drop_columns(df, drop_cols)
        return df

    # This method is used by hsfs_utils to launch a job for python client
    def compute_training_dataset(
        self,
        feature_view_obj,
        user_write_options,
        training_dataset_obj=None,
        training_dataset_version=None,
        spine=None,
        primary_keys=False,
        event_time=False,
        training_helper_columns=False,
    ):
        if training_dataset_obj:
            pass
        elif training_dataset_version:
            training_dataset_obj = self._get_training_dataset_metadata(
                feature_view_obj, training_dataset_version
            )
        else:
            raise ValueError("No training dataset object or version is provided")

        batch_query = self.get_batch_query(
            feature_view_obj,
            training_dataset_obj.event_start_time,
            training_dataset_obj.event_end_time,
            with_label=True,
            primary_keys=primary_keys,
            event_time=event_time,
            inference_helper_columns=False,
            training_helper_columns=training_helper_columns,
            training_dataset_version=training_dataset_obj.version,
            spine=spine,
        )

        # for spark job
        user_write_options["training_helper_columns"] = training_helper_columns
        user_write_options["primary_keys"] = primary_keys
        user_write_options["event_time"] = event_time

        td_job = engine.get_instance().write_training_dataset(
            training_dataset_obj,
            batch_query,
            user_write_options,
            self._OVERWRITE,
            feature_view_obj=feature_view_obj,
        )

        # Set training dataset schema after training dataset has been generated
        training_dataset_obj.schema = self.get_training_dataset_schema(
            feature_view=feature_view_obj,
            training_dataset_version=training_dataset_obj.version,
        )

        if engine.get_type().startswith("spark"):
            # if spark engine, read td and compute stats
            if training_dataset_obj.splits:
                td_df = dict(
                    [
                        (
                            split.name,
                            self._training_dataset_engine.read(
                                training_dataset_obj, split.name, {}
                            ),
                        )
                        for split in training_dataset_obj.splits
                    ]
                )
            else:
                td_df = self._training_dataset_engine.read(
                    training_dataset_obj, None, {}
                )
            self.compute_training_dataset_statistics(
                feature_view_obj, training_dataset_obj, td_df
            )

        return td_job

    def compute_training_dataset_statistics(
        self, feature_view_obj, training_dataset_obj, td_df
    ):
        if training_dataset_obj.statistics_config.enabled:
            if training_dataset_obj.splits:
                if not isinstance(td_df, dict):
                    raise ValueError(
                        "Provided dataframes should be in dict format "
                        "'split': dataframe"
                    )
                return self._statistics_engine.compute_and_save_split_statistics(
                    training_dataset_obj,
                    feature_dataframes=td_df,
                    feature_view_obj=feature_view_obj,
                )
            else:
                return self._statistics_engine.compute_and_save_statistics(
                    training_dataset_obj,
                    feature_dataframe=td_df,
                    feature_view_obj=feature_view_obj,
                )

    def _get_training_dataset_metadata(
        self, feature_view_obj: feature_view.FeatureView, training_dataset_version
    ):
        td = self._feature_view_api.get_training_dataset_by_version(
            feature_view_obj.name, feature_view_obj.version, training_dataset_version
        )

        return td

    def _get_training_datasets_metadata(
        self, feature_view_obj: feature_view.FeatureView
    ):
        tds = self._feature_view_api.get_training_datasets(
            feature_view_obj.name, feature_view_obj.version
        )
        # schema needs to be set for writing training data or feature serving
        for td in tds:
            td.schema = feature_view_obj.get_training_dataset_schema(td.version)
        return tds

    def get_training_datasets(self, feature_view_obj):
        tds = self._get_training_datasets_metadata(feature_view_obj)
        # this is the only place we expose training dataset metadata
        # we return training dataset base classes with metadata only
        return [super(td.__class__, td) for td in tds]

    def _create_training_data_metadata(self, feature_view_obj, training_dataset_obj):
        td = self._feature_view_api.create_training_dataset(
            feature_view_obj.name, feature_view_obj.version, training_dataset_obj
        )
        return td

    def delete_training_data(self, feature_view_obj, training_data_version=None):
        if training_data_version:
            self._feature_view_api.delete_training_data_version(
                feature_view_obj.name, feature_view_obj.version, training_data_version
            )
        else:
            self._feature_view_api.delete_training_data(
                feature_view_obj.name, feature_view_obj.version
            )

    def delete_training_dataset_only(
        self, feature_view_obj, training_data_version=None
    ):
        if training_data_version:
            self._feature_view_api.delete_training_dataset_only_version(
                feature_view_obj.name, feature_view_obj.version, training_data_version
            )
        else:
            self._feature_view_api.delete_training_dataset_only(
                feature_view_obj.name, feature_view_obj.version
            )

    def get_batch_data(
        self,
        feature_view_obj,
        start_time,
        end_time,
        training_dataset_version,
        transformation_functions,
        read_options=None,
        spine=None,
        primary_keys=False,
        event_time=False,
        inference_helper_columns=False,
        dataframe_type="default",
        transformed=True,
    ):
        self._check_feature_group_accessibility(feature_view_obj)

        # check if primary_keys/event_time are ambiguous
        if primary_keys:
            self._get_primary_keys_from_query(feature_view_obj.query)
        if event_time:
            self._get_eventtimes_from_query(feature_view_obj.query)

        feature_dataframe = self.get_batch_query(
            feature_view_obj,
            start_time,
            end_time,
            with_label=False,
            primary_keys=primary_keys,
            event_time=event_time,
            inference_helper_columns=inference_helper_columns or not transformed,
            training_helper_columns=False,
            training_dataset_version=training_dataset_version,
            spine=spine,
        ).read(read_options=read_options, dataframe_type=dataframe_type)
        if transformation_functions and transformed:
            return engine.get_instance()._apply_transformation_function(
                transformation_functions, dataset=feature_dataframe
            )
        else:
            return feature_dataframe

    def transform_batch_data(self, features, transformation_functions):
        return engine.get_instance()._apply_transformation_function(
            transformation_functions, dataset=features, inplace=False
        )

    def add_tag(
        self, feature_view_obj, name: str, value, training_dataset_version=None
    ):
        self._tags_api.add(
            feature_view_obj,
            name,
            value,
            training_dataset_version=training_dataset_version,
        )

    def delete_tag(self, feature_view_obj, name: str, training_dataset_version=None):
        self._tags_api.delete(
            feature_view_obj, name, training_dataset_version=training_dataset_version
        )

    def get_tag(self, feature_view_obj, name: str, training_dataset_version=None):
        return self._tags_api.get(
            feature_view_obj, name, training_dataset_version=training_dataset_version
        )[name]

    def get_tags(self, feature_view_obj, training_dataset_version=None):
        return self._tags_api.get(
            feature_view_obj, training_dataset_version=training_dataset_version
        )

    def get_parent_feature_groups(self, feature_view_obj):
        """Get the parents of this feature view, based on explicit provenance.
        Parents are feature groups or external feature groups. These feature
        groups can be accessible, deleted or inaccessible.
        For deleted and inaccessible feature groups, only a minimal information is
        returned.

        # Arguments
            feature_view_obj: Metadata object of feature view.

        # Returns
            `ProvenanceLinks`:  the feature groups used to generated this feature view
        """
        return self._feature_view_api.get_parent_feature_groups(
            feature_view_obj.name, feature_view_obj.version
        )

    def get_models_provenance(
        self, feature_view_obj, training_dataset_version: Optional[int] = None
    ):
        """Get the generated models using this feature view, based on explicit
        provenance. These models can be accessible or inaccessible. Explicit
        provenance does not track deleted generated model links, so deleted
        will always be empty.
        For inaccessible models, only a minimal information is returned.

        # Arguments
            feature_view_obj: Filter generated models based on feature view (name, version).
            training_dataset_version: Filter generated models based on the used training dataset version.

        # Returns
            `ProvenanceLinks`:  the models generated using this feature group
        """
        return self._feature_view_api.get_models_provenance(
            feature_view_obj.name,
            feature_view_obj.version,
            training_dataset_version=training_dataset_version,
        )

    def _check_feature_group_accessibility(self, feature_view_obj):
        if engine.get_type() == "python":
            try:
                from hsfs.core import arrow_flight_client

                arrow_flight_client_imported = True
            except ImportError:
                arrow_flight_client_imported = False

            if (
                arrow_flight_client_imported
                and arrow_flight_client.get_instance().is_enabled()
            ):
                if not arrow_flight_client.supports(
                    feature_view_obj.query.featuregroups
                ):
                    raise NotImplementedError(
                        "Hopsworks Feature Query Service can only read from cached feature groups"
                        " and external feature groups on BigQuery and Snowflake."
                        " When using other external feature groups please use "
                        "`feature_view.create_training_data` instead. "
                        "If you are using spines, use a Spark Kernel."
                    )
            elif not feature_view_obj.query.is_cache_feature_group_only():
                raise NotImplementedError(
                    "Python kernel can only read from cached feature groups."
                    " When using external feature groups please use "
                    "`feature_view.create_training_data` instead. "
                    "If you are using spines, use a Spark Kernel."
                )

    def _get_feature_view_url(self, fv: "feature_view.FeatureView"):
        path = (
            "/p/"
            + str(client.get_instance()._project_id)
            + "/fs/"
            + str(fv.featurestore_id)
            + "/fv/"
            + str(fv.name)
            + "/version/"
            + str(fv.version)
        )
        return util.get_hostname_replaced_url(path)

    def _get_primary_keys_from_query(self, fv_query_obj, check_duplicate=True):
        fv_pks = set(
            [
                feature.name
                for feature in fv_query_obj._left_feature_group.features
                if feature.primary
            ]
        )
        for _join in fv_query_obj._joins:
            fv_pks.update(
                [
                    (
                        self._check_if_exists_with_prefix(feature.name, fv_pks)
                        if check_duplicate
                        else feature.name
                    )
                    if _join.prefix is None
                    else _join.prefix + feature.name
                    for feature in _join.query._left_feature_group.features
                    if feature.primary
                ]
            )

        return list(fv_pks)

    def _get_eventtimes_from_query(self, fv_query_obj, check_duplicate=True):
        fv_events = set()
        if fv_query_obj._left_feature_group.event_time:
            fv_events.update([fv_query_obj._left_feature_group.event_time])
        for _join in fv_query_obj._joins:
            if _join.query._left_feature_group.event_time:
                fv_events.update(
                    [
                        (
                            self._check_if_exists_with_prefix(
                                _join.query._left_feature_group.event_time, fv_events
                            )
                            if check_duplicate
                            else _join.query._left_feature_group.event_time
                        )
                        if _join.prefix is None
                        else _join.prefix + _join.query._left_feature_group.event_time
                    ]
                )

        return list(fv_events)

    def _check_if_exists_with_prefix(self, f_name, f_set):
        if f_name in f_set:
            raise FeatureStoreException(
                f"Provided feature {f_name} is ambiguous and exists in more than one feature groups."
                "To avoid this error specify prefix in the join."
            )
        else:
            return f_name

    def enable_feature_logging(self, fv):
        self._feature_view_api.enable_feature_logging(fv.name, fv.version)
        fv.logging_enabled = True
        return fv

    def get_feature_logging(self, fv):
        return FeatureLogging.from_response_json(
            self._feature_view_api.get_feature_logging(fv.name, fv.version)
        )

    def _get_logging_fg(self, fv, transformed):
        return self.get_feature_logging(fv).get_feature_group(transformed)

    def log_features(
        self,
        fv: feature_view.FeatureView,
        feature_logging: FeatureLogging,
        untransformed_features: Union[
            pd.DataFrame, list[list], np.ndarray, TypeVar("pyspark.sql.DataFrame")
        ] = None,
        transformed_features: Union[
            pd.DataFrame, list[list], np.ndarray, TypeVar("pyspark.sql.DataFrame")
        ] = None,
        predictions: Optional[Union[pd.DataFrame, list[list], np.ndarray]] = None,
        write_options: Optional[Dict[str, Any]] = None,
        training_dataset_version: Optional[int] = None,
        hsml_model=None,
        logger: FeatureLogger = None,
    ):
        if untransformed_features is None and transformed_features is None:
            return

        default_write_options = {
            "start_offline_materialization": False,
        }
        if write_options:
            default_write_options.update(write_options)
        results = []
        if logger:
            logger.log(
                **{
                    key: (
                        self._get_feature_logging_data(
                            features_rows=features,
                            feature_logging=feature_logging,
                            transformed=transformed,
                            fv=fv,
                            predictions=predictions,
                            training_dataset_version=training_dataset_version,
                            hsml_model=hsml_model,
                            return_list=True,
                        )
                        if features
                        else None
                    )
                    for transformed, key, features in [
                        (False, "untransformed_features", untransformed_features),
                        (True, "transformed_features", transformed_features),
                    ]
                }
            )

        else:
            for transformed, features in [
                (False, untransformed_features),
                (True, transformed_features),
            ]:
                fg = feature_logging.get_feature_group(transformed)
                if features is None:
                    continue
                results.append(
                    fg.insert(
                        self._get_feature_logging_data(
                            features_rows=features,
                            feature_logging=feature_logging,
                            transformed=transformed,
                            fv=fv,
                            predictions=predictions,
                            training_dataset_version=training_dataset_version,
                            hsml_model=hsml_model,
                            return_list=False,
                        ),
                        write_options=default_write_options,
                    )
                )
        return results

    def _get_feature_logging_data(
        self,
        features_rows,
        feature_logging,
        transformed,
        fv,
        predictions,
        training_dataset_version,
        hsml_model,
        return_list=False,
    ):
        fg = feature_logging.get_feature_group(transformed)
        training_dataset_schema = fv.get_training_dataset_schema(
            training_dataset_version=training_dataset_version
        )
        td_predictions = [
            feature for feature in training_dataset_schema if feature.label
        ]
        td_predictions_names = set([feature.name for feature in td_predictions])
        if transformed:
            td_features = [
                feature.name
                for feature in training_dataset_schema
                if feature.name not in td_predictions_names
            ]
        else:
            td_features = [
                feature.name
                for feature in fv.features
                if feature.name not in td_predictions_names
            ]

        if return_list:
            return engine.get_instance().get_feature_logging_list(
                features_rows,
                fg=fg,
                td_features=td_features,
                td_predictions=td_predictions,
                td_col_name=FeatureViewEngine._LOG_TD_VERSION,
                time_col_name=FeatureViewEngine._LOG_TIME,
                model_col_name=FeatureViewEngine._HSML_MODEL,
                predictions=predictions,
                training_dataset_version=training_dataset_version,
                hsml_model=self.get_hsml_model_value(hsml_model)
                if hsml_model
                else None,
            )
        else:
            return engine.get_instance().get_feature_logging_df(
                features_rows,
                fg=fg,
                td_features=td_features,
                td_predictions=td_predictions,
                td_col_name=FeatureViewEngine._LOG_TD_VERSION,
                time_col_name=FeatureViewEngine._LOG_TIME,
                model_col_name=FeatureViewEngine._HSML_MODEL,
                predictions=predictions,
                training_dataset_version=training_dataset_version,
                hsml_model=self.get_hsml_model_value(hsml_model)
                if hsml_model
                else None,
            )

    def read_feature_logs(
        self,
        fv,
        start_time: Optional[Union[str, int, datetime, datetime.date]] = None,
        end_time: Optional[Union[str, int, datetime, datetime.date]] = None,
        filter: Optional[Union[Filter, Logic]] = None,
        transformed: Optional[bool] = False,
        training_dataset_version=None,
        hsml_model=None,
    ):
        fg = self._get_logging_fg(fv, transformed)
        fv_feat_name_map = self._get_fv_feature_name_map(fv)
        query = fg.select_all()
        if start_time:
            query = query.filter(
                fg.get_feature(FeatureViewEngine._LOG_TIME) >= start_time
            )
        if end_time:
            query = query.filter(
                fg.get_feature(FeatureViewEngine._LOG_TIME) <= end_time
            )
        if training_dataset_version:
            query = query.filter(
                fg.get_feature(FeatureViewEngine._LOG_TD_VERSION)
                == training_dataset_version
            )
        if hsml_model:
            query = query.filter(
                fg.get_feature(FeatureViewEngine._HSML_MODEL)
                == self.get_hsml_model_value(hsml_model)
            )
        if filter:
            query = query.filter(
                self._convert_to_log_fg_filter(fg, fv, filter, fv_feat_name_map)
            )
        return engine.get_instance().read_feature_log(query, self._LOG_TIME)

    @staticmethod
    def get_hsml_model_value(hsml_model):
        return f"{hsml_model.name}_{hsml_model.version}"

    def _convert_to_log_fg_filter(self, fg, fv, filter, fv_feat_name_map):
        if filter is None:
            return None

        if isinstance(filter, Logic):
            return Logic(
                filter.type,
                left_f=self._convert_to_log_fg_filter(fv, filter.left_f),
                right_f=self._convert_to_log_fg_filter(fv, filter.right_f),
                left_l=self._convert_to_log_fg_filter(fv, filter.left_l),
                right_l=self._convert_to_log_fg_filter(fv, filter.right_l),
            )
        elif isinstance(filter, Filter):
            fv_feature_name = fv_feat_name_map.get(
                f"{filter.feature.feature_group_id}_{filter.feature.name}"
            )
            if fv_feature_name is None:
                raise FeatureStoreException(
                    "Filter feature {filter.feature.name} does not exist in feature view feature."
                )
            return Filter(
                fg.get_feature(filter.feature.name),
                filter.condition,
                filter.value,
            )
        else:
            raise FeatureStoreException("Accept only Filter or Logic")

    def _get_fv_feature_name_map(self, fv) -> Dict[str, str]:
        result_dict = {}
        for td_feature in fv.features:
            fg_feature_key = (
                f"{td_feature.feature_group.id}_{td_feature.feature_group_feature_name}"
            )
            result_dict[fg_feature_key] = td_feature.name
        return result_dict

    def get_log_timeline(
        self,
        fv,
        wallclock_time: Optional[Union[str, int, datetime, datetime.date]] = None,
        limit: Optional[int] = None,
        transformed: Optional[bool] = False,
    ) -> Dict[str, Dict[str, str]]:
        fg = self._get_logging_fg(fv, transformed)
        return fg.commit_details(wallclock_time=wallclock_time, limit=limit)

    def pause_logging(self, fv):
        self._feature_view_api.pause_feature_logging(fv.name, fv.version)

    def resume_logging(self, fv):
        self._feature_view_api.resume_feature_logging(fv.name, fv.version)

    def materialize_feature_logs(self, fv, wait, transform):
        if transform is None:
            jobs = [
                self._get_logging_fg(fv, True).materialization_job,
                self._get_logging_fg(fv, False).materialization_job,
            ]
        else:
            jobs = [self._get_logging_fg(fv, transform).materialization_job]
        for job in jobs:
            job.run(await_termination=False)
        if wait:
            for job in jobs:
                job._wait_for_job(wait)
        return jobs

    def delete_feature_logs(self, fv, feature_logging, transformed):
        self._feature_view_api.delete_feature_logs(fv.name, fv.version, transformed)
        feature_logging.update(self.get_feature_logging(fv))



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
from typing import Dict, List, Union

from hsml import (
    client,
    deployable_component_logs,
    deployment,
    inference_endpoint,
    predictor_state,
)
from hsml.client.istio.utils.infer_type import (
    InferInput,
    InferOutput,
    InferRequest,
)
from hsml.constants import ARTIFACT_VERSION
from hsml.constants import INFERENCE_ENDPOINTS as IE


class ServingApi:
    def __init__(self):
        pass

    def get_by_id(self, id: int):
        """Get the metadata of a deployment with a certain id.

        :param id: id of the deployment
        :type id: int
        :return: deployment metadata object
        :rtype: Deployment
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "serving",
            str(id),
        ]
        deployment_json = _client._send_request("GET", path_params)
        deployment_instance = deployment.Deployment.from_response_json(deployment_json)
        deployment_instance.model_registry_id = _client._project_id
        return deployment_instance

    def get(self, name: str):
        """Get the metadata of a deployment with a certain name.

        :param name: name of the deployment
        :type name: str
        :return: deployment metadata object
        :rtype: Deployment
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "serving"]
        query_params = {"name": name}
        deployment_json = _client._send_request(
            "GET", path_params, query_params=query_params
        )
        deployment_instance = deployment.Deployment.from_response_json(deployment_json)
        deployment_instance.model_registry_id = _client._project_id
        return deployment_instance

    def get_all(self, model_name: str = None, status: str = None):
        """Get the metadata of all deployments.

        :return: model metadata objects
        :rtype: List[Deployment]
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "serving"]
        query_params = {
            "model": model_name,
            "status": status.capitalize() if status is not None else None,
        }
        deployments_json = _client._send_request(
            "GET", path_params, query_params=query_params
        )
        deployment_instances = deployment.Deployment.from_response_json(
            deployments_json
        )
        for deployment_instance in deployment_instances:
            deployment_instance.model_registry_id = _client._project_id
        return deployment_instances

    def get_inference_endpoints(self):
        """Get inference endpoints.

        :return: inference endpoints for the current project.
        :rtype: List[InferenceEndpoint]
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "inference", "endpoints"]
        endpoints_json = _client._send_request("GET", path_params)
        return inference_endpoint.InferenceEndpoint.from_response_json(endpoints_json)

    def put(self, deployment_instance):
        """Save deployment metadata to model serving.

        :param deployment_instance: metadata object of deployment to be saved
        :type deployment_instance: Deployment
        :return: updated metadata object of the deployment
        :rtype: Deployment
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "serving"]
        headers = {"content-type": "application/json"}

        if deployment_instance.artifact_version == ARTIFACT_VERSION.CREATE:
            deployment_instance.artifact_version = -1

        deployment_instance = deployment_instance.update_from_response_json(
            _client._send_request(
                "PUT",
                path_params,
                headers=headers,
                data=deployment_instance.json(),
            )
        )
        deployment_instance.model_registry_id = _client._project_id
        return deployment_instance

    def post(self, deployment_instance, action: str):
        """Perform an action on the deployment

        :param action: action to perform on the deployment (i.e., START or STOP)
        :type action: str
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "serving",
            deployment_instance.id,
        ]
        query_params = {"action": action}
        return _client._send_request("POST", path_params, query_params=query_params)

    def delete(self, deployment_instance):
        """Delete the deployment and metadata.

        :param deployment_instance: metadata object of the deployment to delete
        :type deployment_instance: Deployment
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "serving",
            deployment_instance.id,
        ]
        return _client._send_request("DELETE", path_params)

    def get_state(self, deployment_instance):
        """Get the state of a given deployment

        :param deployment_instance: metadata object of the deployment to get state of
        :type deployment_instance: Deployment
        :return: predictor state
        :rtype: PredictorState
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "serving",
            str(deployment_instance.id),
        ]
        deployment_json = _client._send_request("GET", path_params)
        return predictor_state.PredictorState.from_response_json(deployment_json)

    def reset_changes(self, deployment_instance):
        """Reset a given deployment to the original values in the Hopsworks instance

        :param deployment_instance: metadata object of the deployment to reset
        :type deployment_instance: Deployment
        :return: deployment with reset values
        :rtype: Deployment
        """

        _client = client.get_instance()
        path_params = ["project", _client._project_id, "serving"]
        query_params = {"name": deployment_instance.name}
        deployment_json = _client._send_request(
            "GET", path_params, query_params=query_params
        )
        deployment_aux = deployment_instance.update_from_response_json(deployment_json)
        # TODO: remove when model_registry_id is added properly to deployments in backend
        deployment_aux.model_registry_id = _client._project_id
        return deployment_aux

    def send_inference_request(
        self,
        deployment_instance,
        data: Union[Dict, List[InferInput]],
        through_hopsworks: bool = False,
    ) -> Union[Dict, List[InferOutput]]:
        """Send inference requests to a deployment with a certain id

        :param deployment_instance: metadata object of the deployment to be used for the prediction
        :type deployment_instance: Deployment
        :param data: payload of the inference request
        :type data: Union[Dict, List[InferInput]]
        :param through_hopsworks: whether to send the inference request through the Hopsworks REST API or not
        :type through_hopsworks: bool
        :return: inference response
        :rtype: Union[Dict, List[InferOutput]]
        """
        if deployment_instance.api_protocol == IE.API_PROTOCOL_REST:
            # REST protocol, use hopsworks or istio client
            return self._send_inference_request_via_rest_protocol(
                deployment_instance, data, through_hopsworks
            )
        else:
            # gRPC protocol, use the deployment grpc channel
            return self._send_inference_request_via_grpc_protocol(
                deployment_instance, data
            )

    def _send_inference_request_via_rest_protocol(
        self,
        deployment_instance,
        data: Dict,
        through_hopsworks: bool = False,
    ) -> Dict:
        headers = {"content-type": "application/json"}
        if through_hopsworks:
            # use Hopsworks client
            _client = client.get_instance()
            path_params = self._get_hopsworks_inference_path(
                _client._project_id, deployment_instance
            )
            with_base_path_params = True
        else:
            _client = client.istio.get_instance()
            if _client is not None:
                # use istio client
                path_params = self._get_istio_inference_path(deployment_instance)
                # - add host header
                headers["host"] = self._get_inference_request_host_header(
                    deployment_instance.project_namespace,
                    deployment_instance.name,
                    client.get_knative_domain(),
                )
                with_base_path_params = False
            else:
                # fallback to Hopsworks client
                _client = client.get_instance()
                path_params = self._get_hopsworks_inference_path(
                    _client._project_id, deployment_instance
                )
                with_base_path_params = True

        # send inference request
        return _client._send_request(
            "POST",
            path_params,
            headers=headers,
            data=json.dumps(data),
            with_base_path_params=with_base_path_params,
        )

    def _send_inference_request_via_grpc_protocol(
        self, deployment_instance, data: List[InferInput]
    ) -> List[InferOutput]:
        # get grpc channel
        if deployment_instance._grpc_channel is None:
            # The gRPC channel is lazily initialized. The first call to deployment.predict() will initialize
            # the channel, which will be reused in all following calls on the same deployment object.
            # The gRPC channel is freed when calling deployment.stop()
            print("Initializing gRPC channel...")
            deployment_instance._grpc_channel = self._create_grpc_channel(
                deployment_instance
            )
        # build an infer request
        request = InferRequest(
            infer_inputs=data,
            model_name=deployment_instance.name,
        )

        # send infer request
        infer_response = deployment_instance._grpc_channel.infer(
            infer_request=request, headers=None
        )

        # extract infer outputs
        return infer_response.outputs

    def _create_grpc_channel(self, deployment_instance):
        _client = client.istio.get_instance()
        service_hostname = self._get_inference_request_host_header(
            deployment_instance.project_namespace,
            deployment_instance.name,
            client.get_knative_domain(),
        )
        return _client._create_grpc_channel(service_hostname)

    def is_kserve_installed(self):
        """Check if kserve is installed

        :return: whether kserve is installed
        :rtype: bool
        """

        _client = client.get_instance()
        path_params = ["variables", "kube_kserve_installed"]
        kserve_installed = _client._send_request("GET", path_params)
        return (
            "successMessage" in kserve_installed
            and kserve_installed["successMessage"] == "true"
        )

    def get_resource_limits(self):
        """Get resource limits for model serving"""

        _client = client.get_instance()

        path_params = ["variables", "kube_serving_max_cores_allocation"]
        max_cores = _client._send_request("GET", path_params)

        path_params = ["variables", "kube_serving_max_memory_allocation"]
        max_memory = _client._send_request("GET", path_params)

        path_params = ["variables", "kube_serving_max_gpus_allocation"]
        max_gpus = _client._send_request("GET", path_params)

        return {
            "cores": float(max_cores["successMessage"]),
            "memory": int(max_memory["successMessage"]),
            "gpus": int(max_gpus["successMessage"]),
        }

    def get_num_instances_limits(self):
        """Get number of instances limits for model serving"""

        _client = client.get_instance()

        path_params = ["variables", "kube_serving_min_num_instances"]
        min_instances = _client._send_request("GET", path_params)

        path_params = ["variables", "kube_serving_max_num_instances"]
        max_instances = _client._send_request("GET", path_params)

        return [
            int(min_instances["successMessage"]),
            int(max_instances["successMessage"]),
        ]

    def get_knative_domain(self):
        """Get the domain used by knative"""

        _client = client.get_instance()

        path_params = ["variables", "kube_knative_domain_name"]
        domain = _client._send_request("GET", path_params)

        return domain["successMessage"]

    def get_logs(self, deployment_instance, component, tail):
        """Get the logs of a deployment

        :param deployment_instance: metadata object of the deployment to get logs from
        :type deployment_instance: Deployment
        :param component: deployment component (e.g., predictor or transformer)
        :type component: str
        :param tail: number of tailing lines to retrieve
        :type tail: int
        :return: deployment logs
        :rtype: DeployableComponentLogs
        """

        _client = client.get_instance()
        path_params = [
            "project",
            _client._project_id,
            "serving",
            deployment_instance.id,
            "logs",
        ]
        query_params = {"component": component, "tail": tail}
        return deployable_component_logs.DeployableComponentLogs.from_response_json(
            _client._send_request("GET", path_params, query_params=query_params)
        )

    def _get_inference_request_host_header(
        self, project_namespace: str, deployment_name: str, domain: str
    ):
        return "{}.{}.{}".format(deployment_name, project_namespace, domain).lower()

    def _get_hopsworks_inference_path(self, project_id: int, deployment_instance):
        return [
            "project",
            project_id,
            "inference",
            "models",
            deployment_instance.name + ":predict",
        ]

    def _get_istio_inference_path(self, deployment_instance):
        if deployment_instance.model_server == "VLLM":
            return ["openai", "v1", "completions"]

        return ["v1", "models", deployment_instance.name + ":predict"]



#
#   Copyright 2020 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

from __future__ import annotations

import base64
import logging
import os

import hopsworks_common.client
import requests
from hopsworks_common.client import auth, base, exceptions
from hopsworks_common.client.exceptions import FeatureStoreException


try:
    from pyspark.sql import SparkSession
except ImportError:
    pass


_logger = logging.getLogger(__name__)


class Client(base.Client):
    def __init__(
        self,
        host,
        port,
        project,
        engine,
        hostname_verification,
        trust_store_path,
        cert_folder,
        api_key_file,
        api_key_value,
    ):
        """Initializes a client in an external environment such as AWS Sagemaker."""
        _logger.info("Initializing external client")
        if not host:
            raise exceptions.ExternalClientError("host")

        self._host = host
        self._port = port
        self._base_url = "https://" + self._host + ":" + str(self._port)
        _logger.info("Base URL: %s", self._base_url)

        if api_key_value is not None:
            _logger.debug("Using provided API key value")
            api_key = api_key_value
        else:
            _logger.debug(f"Reading api key from {api_key_file}")
            with open(api_key_file) as f:
                api_key = f.readline().strip()

        _logger.debug("Using api key to setup header authentification")
        self._auth = auth.ApiKeyAuth(api_key)

        _logger.debug("Setting up requests session")
        self._session = requests.session()
        self._connected = True

        self._verify = self._get_verify(hostname_verification, trust_store_path)
        self._hostname_verification = hostname_verification
        self._hopsworks_ca_trust_store_path = trust_store_path
        _logger.debug("Verify: %s", self._verify)

        self._cert_key = None
        self._cert_folder_base = cert_folder
        self._cert_folder = None

        self._engine = engine

        if not project:
            # This should only happen due to the way of work of and in hopsworks.login:
            # It first initalizes a client, and then uses it to figure out which project to use.
            self._project_name = None
            return

        self.provide_project(project)

    def provide_project(self, project):
        self._project_name = project
        _logger.debug("Project name: %s", self._project_name)

        project_info = self._get_project_info(project)
        self._project_id = str(project_info["projectId"])
        _logger.debug("Setting Project ID: %s", self._project_id)

        if self._engine == "python":
            self.download_certs()

        elif self._engine == "spark":
            # When using the Spark engine with metastore connection, the certificates
            # are needed when the application starts (before user code is run)
            # So in this case, we can't materialize the certificates on the fly.
            _logger.debug("Running in Spark environment, initializing Spark session")
            _spark_session = SparkSession.builder.enableHiveSupport().getOrCreate()

            self._validate_spark_configuration(_spark_session)
            with open(
                _spark_session.conf.get("spark.hadoop.hops.ssl.keystores.passwd.name"),
                "r",
            ) as f:
                self._cert_key = f.read()

            self._trust_store_path = _spark_session.conf.get(
                "spark.hadoop.hops.ssl.trustore.name"
            )
            self._key_store_path = _spark_session.conf.get(
                "spark.hadoop.hops.ssl.keystore.name"
            )

        elif self._engine == "spark-no-metastore":
            _logger.debug(
                "Running in Spark environment with no metastore, initializing Spark session"
            )
            _spark_session = SparkSession.builder.getOrCreate()
            self.download_certs()

            # Set credentials location in the Spark configuration
            # Set other options in the Spark configuration
            configuration_dict = {
                "hops.ssl.trustore.name": self._trust_store_path,
                "hops.ssl.keystore.name": self._key_store_path,
                "hops.ssl.keystores.passwd.name": self._cert_key_path,
                "fs.permissions.umask-mode": "0002",
                "fs.hopsfs.impl": "io.hops.hopsfs.client.HopsFileSystem",
                "hops.rpc.socket.factory.class.default": "io.hops.hadoop.shaded.org.apache.hadoop.net.HopsSSLSocketFactory",
                "client.rpc.ssl.enabled.protocol": "TLSv1.2",
                "hops.ssl.hostname.verifier": "ALLOW_ALL",
                "hops.ipc.server.ssl.enabled": "true",
            }

            for conf_key, conf_value in configuration_dict.items():
                _spark_session._jsc.hadoopConfiguration().set(conf_key, conf_value)

        hopsworks_common.client.get_connection()._provide_project()

    def download_certs(self):
        res = self._materialize_certs()
        self._write_pem_file(res["caChain"], self._get_ca_chain_path())
        self._write_pem_file(res["clientCert"], self._get_client_cert_path())
        self._write_pem_file(res["clientKey"], self._get_client_key_path())
        return res

    def _materialize_certs(self):
        self._cert_folder = os.path.join(
            self._cert_folder_base, self._host, self._project_name
        )
        self._trust_store_path = os.path.join(self._cert_folder, "trustStore.jks")
        self._key_store_path = os.path.join(self._cert_folder, "keyStore.jks")

        if os.path.exists(self._cert_folder):
            _logger.debug(
                f"Running in Python environment, reading certificates from certificates folder {self._cert_folder_base}"
            )
            _logger.debug("Found certificates: %s", os.listdir(self._cert_folder_base))
        else:
            _logger.debug(
                f"Running in Python environment, creating certificates folder {self._cert_folder_base}"
            )
            os.makedirs(self._cert_folder, exist_ok=True)

        credentials = self._get_credentials(self._project_id)
        self._write_b64_cert_to_bytes(
            str(credentials["kStore"]),
            path=self._get_jks_key_store_path(),
        )
        self._write_b64_cert_to_bytes(
            str(credentials["tStore"]),
            path=self._get_jks_trust_store_path(),
        )

        self._cert_key = str(credentials["password"])
        self._cert_key_path = os.path.join(self._cert_folder, "material_passwd")
        with open(self._cert_key_path, "w") as f:
            f.write(str(credentials["password"]))

        # Return the credentials object for the Python engine to materialize the pem files.
        return credentials

    def _validate_spark_configuration(self, _spark_session):
        exception_text = "Spark is misconfigured for communication with Hopsworks, missing or invalid property: "

        configuration_dict = {
            "spark.hadoop.hops.ssl.trustore.name": None,
            "spark.hadoop.hops.rpc.socket.factory.class.default": "io.hops.hadoop.shaded.org.apache.hadoop.net.HopsSSLSocketFactory",
            "spark.serializer": "org.apache.spark.serializer.KryoSerializer",
            "spark.hadoop.hops.ssl.hostname.verifier": "ALLOW_ALL",
            "spark.hadoop.hops.ssl.keystore.name": None,
            "spark.hadoop.fs.hopsfs.impl": "io.hops.hopsfs.client.HopsFileSystem",
            "spark.hadoop.hops.ssl.keystores.passwd.name": None,
            "spark.hadoop.hops.ipc.server.ssl.enabled": "true",
            "spark.hadoop.client.rpc.ssl.enabled.protocol": "TLSv1.2",
            "spark.hadoop.hive.metastore.uris": None,
            "spark.sql.hive.metastore.jars": None,
        }
        _logger.debug("Configuration dict: %s", configuration_dict)

        for key, value in configuration_dict.items():
            _logger.debug("Validating key: %s", key)
            if not (
                _spark_session.conf.get(key, "not_found") != "not_found"
                and (value is None or _spark_session.conf.get(key, None) == value)
            ):
                raise FeatureStoreException(exception_text + key)

    def _close(self):
        """Closes a client and deletes certificates."""
        _logger.info("Closing external client and cleaning up certificates.")
        self._connected = False
        if self._cert_folder is None:
            _logger.debug("No certificates to clean up.")
            # On external Spark clients (Databricks, Spark Cluster),
            # certificates need to be provided before the Spark application starts.
            return

        # Clean up only on AWS
        _logger.debug("Cleaning up certificates. AWS only.")
        self._cleanup_file(self._get_jks_key_store_path())
        self._cleanup_file(self._get_jks_trust_store_path())
        self._cleanup_file(os.path.join(self._cert_folder, "material_passwd"))
        self._cleanup_file(self._get_ca_chain_path())
        self._cleanup_file(self._get_client_cert_path())
        self._cleanup_file(self._get_client_key_path())

        try:
            # delete project level
            os.rmdir(self._cert_folder)
            # delete host level
            os.rmdir(os.path.dirname(self._cert_folder))
            # on AWS base dir will be empty, and can be deleted otherwise raises OSError
            os.rmdir(self._cert_folder_base)
        except OSError:
            pass

        self._cert_folder = None

    def _get_jks_trust_store_path(self):
        _logger.debug("Getting trust store path: %s", self._trust_store_path)
        return self._trust_store_path

    def _get_jks_key_store_path(self):
        _logger.debug("Getting key store path: %s", self._key_store_path)
        return self._key_store_path

    def _get_ca_chain_path(self) -> str:
        path = os.path.join(
            self._cert_folder_base, self._host, self._project_name, "ca_chain.pem"
        )
        _logger.debug(f"Getting ca chain path {path}")
        return path

    def _get_client_cert_path(self) -> str:
        path = os.path.join(
            self._cert_folder_base, self._host, self._project_name, "client_cert.pem"
        )
        _logger.debug(f"Getting client cert path {path}")
        return path

    def _get_client_key_path(self) -> str:
        path = os.path.join(
            self._cert_folder_base, self._host, self._project_name, "client_key.pem"
        )
        _logger.debug(f"Getting client key path {path}")
        return path

    def _get_project_info(self, project_name):
        """Makes a REST call to hopsworks to get all metadata of a project for the provided project.

        :param project_name: the name of the project
        :type project_name: str
        :return: JSON response with project info
        :rtype: dict
        """
        _logger.debug("Getting project info for project: %s", project_name)
        return self._send_request("GET", ["project", "getProjectInfo", project_name])

    def _write_b64_cert_to_bytes(self, b64_string, path):
        """Converts b64 encoded certificate to bytes file .

        :param b64_string:  b64 encoded string of certificate
        :type b64_string: str
        :param path: path where file is saved, including file name. e.g. /path/key-store.jks
        :type path: str
        """
        _logger.debug(f"Writing b64 encoded certificate to {path}")
        with open(path, "wb") as f:
            cert_b64 = base64.b64decode(b64_string)
            f.write(cert_b64)

    def _cleanup_file(self, file_path):
        """Removes local files with `file_path`."""
        _logger.debug(f"Cleaning up file {file_path}")
        try:
            os.remove(file_path)
        except OSError:
            pass

    def replace_public_host(self, url):
        """no need to replace as we are already in external client"""
        return url

    def _is_external(self):
        return True

    @property
    def host(self):
        return self._host



#
#   Copyright 2022 Logical Clocks AB
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from __future__ import annotations

import json
from enum import Enum
from typing import Optional, Set

import humps
from hsfs import feature_group, feature_view, storage_connector, training_dataset, util


class Artifact:
    class MetaType(Enum):
        DELETED = 1
        INACCESSIBLE = 2
        FAULTY = 3
        NOT_SUPPORTED = 4

        def json(self):
            return json.dumps(self, cls=util.Encoder)

        def to_dict(self):
            return self.name

        def __str__(self):
            return self.json()

        def __repr__(self):
            return f"<MetaType.{self.name}>"

    def __init__(
        self,
        feature_store_name,
        name,
        version,
        type,
        meta_type,
        href=None,
        exception_cause=None,
        **kwargs,
    ):
        self._feature_store_name = feature_store_name
        self._name = name
        self._version = version
        self._type = type
        self._meta_type = meta_type
        self._href = href
        self._exception_cause = exception_cause

    @property
    def feature_store_name(self):
        """Name of the feature store in which the artifact is located."""
        return self._feature_store_name

    @property
    def name(self):
        """Name of the artifact."""
        return self._name

    @property
    def version(self):
        """Version of the artifact"""
        return self._version

    def json(self):
        return json.dumps(self, cls=util.Encoder)

    def to_dict(self):
        return {
            "feature_store_name": self._feature_store_name,
            "name": self.name,
            "version": self._version,
            "type": self._type,
            "meta_type": self._meta_type,
            "href": self._href,
            "err": self._exception_cause,
        }

    def __str__(self):
        return self.json()

    def __repr__(self):
        return (
            f"Artifact({self._feature_store_name!r}, {self._name!r}, "
            f"{self._version!r}, {self._type!r}, {self._meta_type!r}, "
            f"{self._href!r}, {self._exception_cause!r})"
        )

    @staticmethod
    def from_response_json(json_dict: dict):
        link_json = humps.decamelize(json_dict)
        href = None
        exception_cause = None
        if link_json.get("exception_cause") is not None:
            meta_type = Artifact.MetaType.FAULTY
            exception_cause = link_json.get("exception_cause")
        elif bool(link_json["deleted"]):
            meta_type = Artifact.MetaType.DELETED
        elif not bool(link_json["accessible"]):
            meta_type = Artifact.MetaType.INACCESSIBLE
            href = link_json["artifact"]["href"]
        else:
            meta_type = Artifact.MetaType.NOT_SUPPORTED
            href = link_json["artifact"]["href"]
        return Artifact(
            link_json["artifact"]["project"],
            link_json["artifact"]["name"],
            link_json["artifact"]["version"],
            link_json["artifact_type"],
            meta_type,
            href=href,
            exception_cause=exception_cause,
        )


class Links:
    def __init__(self):
        self._accessible = []
        self._deleted = []
        self._inaccessible = []
        self._faulty = []

    @property
    def deleted(self):
        """List of [Artifact objects] which contains
        minimal information (name, version) about the entities
        (storage connectors, feature groups, feature views, models) they represent.
        These entities have been removed from the feature store/model registry.
        """
        return self._deleted

    @property
    def inaccessible(self):
        """List of [Artifact objects] which contains
        minimal information (name, version) about the entities
        (storage connectors, feature groups, feature views, models) they represent.
        These entities exist in the feature store/model registry, however the user
        does not have access to them anymore.
        """
        return self._inaccessible

    @property
    def accessible(self):
        """List of [StorageConnectors|FeatureGroups|FeatureViews|Models] objects
        which are part of the provenance graph requested. These entities
        exist in the feature store/model registry and the user has access to them.
        """
        return self._accessible

    @property
    def faulty(self):
        """List of [Artifact objects] which contains
        minimal information (name, version) about the entities
        (storage connectors, feature groups, feature views, models) they represent.
        These entities exist in the feature store/model registry, however they are corrupted.
        """
        return self._faulty

    class Direction(Enum):
        UPSTREAM = 1
        DOWNSTREAM = 2

    class Type(Enum):
        FEATURE_GROUP = 1
        FEATURE_VIEW = 2
        MODEL = 3
        STORAGE_CONNECTOR = 4

    def __str__(self, indent=None):
        return json.dumps(self, cls=ProvenanceEncoder, indent=indent)

    def __repr__(self):
        return (
            f"Links({self._accessible!r}, {self._deleted!r}"
            f", {self._inaccessible!r}, {self._faulty!r})"
        )

    @staticmethod
    def __parse_storage_connector(links_json: dict, artifacts: Set[str]):
        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] in artifacts:
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(Artifact.from_response_json(link_json["node"]))
                elif bool(link_json["node"]["accessible"]):
                    links.accessible.append(
                        storage_connector.StorageConnector.from_response_json(
                            link_json["node"]["artifact"]
                        )
                    )
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(Artifact.from_response_json(link_json["node"]))
                else:
                    links.inaccessible.append(
                        Artifact.from_response_json(link_json["node"])
                    )
        return links

    @staticmethod
    def __feature_group(link_json: dict):
        if link_json["artifact_type"] == "FEATURE_GROUP":
            return feature_group.FeatureGroup.from_response_json(link_json["artifact"])
        elif link_json["artifact_type"] == "EXTERNAL_FEATURE_GROUP":
            return feature_group.ExternalFeatureGroup.from_response_json(
                link_json["artifact"]
            )

    @staticmethod
    def __parse_feature_groups(links_json: dict, artifacts: Set[str]):
        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] in artifacts:
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(Artifact.from_response_json(link_json["node"]))
                elif bool(link_json["node"]["accessible"]):
                    links.accessible.append(Links.__feature_group(link_json["node"]))
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(Artifact.from_response_json(link_json["node"]))
                else:
                    links.inaccessible.append(
                        Artifact.from_response_json(link_json["node"])
                    )
        return links

    @staticmethod
    def __parse_feature_views(links_json: dict, artifacts: Set[str]):
        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] in artifacts:
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(Artifact.from_response_json(link_json["node"]))
                elif bool(link_json["node"]["accessible"]):
                    links.accessible.append(
                        feature_view.FeatureView.from_response_json(
                            link_json["node"]["artifact"]
                        )
                    )
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(Artifact.from_response_json(link_json["node"]))
                else:
                    links.inaccessible.append(
                        Artifact.from_response_json(link_json["node"])
                    )
        return links

    @staticmethod
    def __parse_models(
        links_json: dict, training_dataset_version: Optional[int] = None
    ):
        from hsml import model
        from hsml.core import explicit_provenance as hsml_explicit_provenance

        links = Links()
        for link_json in links_json:
            if link_json["node"]["artifact_type"] == "MODEL":
                if link_json["node"].get("exception_cause") is not None:
                    links._faulty.append(
                        hsml_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
                elif bool(link_json["node"]["accessible"]):
                    links.accessible.append(
                        model.Model.from_response_json(link_json["node"]["artifact"])
                    )
                elif bool(link_json["node"]["deleted"]):
                    links.deleted.append(
                        hsml_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
                else:
                    links.inaccessible.append(
                        hsml_explicit_provenance.Artifact.from_response_json(
                            link_json["node"]
                        )
                    )
            else:
                # the only artifact types here are MODEL and TRAINING_DATASET
                # elif link_json["node"]["artifact_type"] == "TRAINING_DATASET":
                if (
                    training_dataset_version
                    and link_json["node"]["artifact"]["version"]
                    != training_dataset_version
                ):
                    # Skip the following operations if the versions don't match
                    pass
                else:
                    new_links = Links.__parse_models(link_json["downstream"])
                    links.faulty.extend(new_links.faulty)
                    links.accessible.extend(new_links.accessible)
                    links.inaccessible.extend(new_links.inaccessible)
                    links.deleted.extend(new_links.deleted)

        return links

    @staticmethod
    def from_response_json(
        json_dict: dict,
        direction: Direction,
        artifact: Type,
        training_dataset_version: Optional[int] = None,
    ):
        """Parse explicit links from json response. There are four types of
        Links: UpstreamStorageConnectors, UpstreamFeatureGroups, DownstreamFeatureGroups, DownstreamFeatureViews

        # Arguments
            json_dict: json response from the explicit provenance endpoint
            direction: subset of links to parse - UPSTREAM/DOWNSTREAM
            artifact: subset of links to parse - STORAGE_CONNECTOR/FEATURE_GROUP/FEATURE_VIEW/MODEL
            training_dataset_version: training dataset version

        # Returns
            A ProvenanceLink object for the selected parse type.
        """
        links_json = humps.decamelize(json_dict)

        if direction == Links.Direction.DOWNSTREAM and artifact == Links.Type.MODEL:
            import importlib.util

            if not importlib.util.find_spec("hsml"):
                raise Exception(
                    "hsml is not installed in the environment - cannot parse model registry artifacts"
                )
            if not importlib.util.find_spec("hopsworks"):
                raise Exception(
                    "hopsworks is not installed in the environment - cannot switch from hsml connection to hsfs connection"
                )

            # make sure the hsml connection is initialized so that the model can actually be used after being returned
            import hopsworks

            if not hopsworks._connected_project:
                raise Exception(
                    "hopsworks connection is not initialized - use hopsworks.login to connect if you want the ability to use provenance with connections between hsfs and hsml"
                )

            hopsworks._connected_project.get_model_registry()

            return Links.__parse_models(
                links_json["downstream"],
                training_dataset_version=training_dataset_version,
            )
        else:
            if direction == Links.Direction.UPSTREAM:
                if artifact == Links.Type.FEATURE_GROUP:
                    return Links.__parse_feature_groups(
                        links_json["upstream"],
                        {
                            "FEATURE_GROUP",
                            "EXTERNAL_FEATURE_GROUP",
                        },
                    )
                elif artifact == Links.Type.STORAGE_CONNECTOR:
                    return Links.__parse_storage_connector(
                        links_json["upstream"], {"STORAGE_CONNECTOR"}
                    )
                else:
                    return Links()

            if direction == Links.Direction.DOWNSTREAM:
                if artifact == Links.Type.FEATURE_GROUP:
                    return Links.__parse_feature_groups(
                        links_json["downstream"],
                        {
                            "FEATURE_GROUP",
                            "EXTERNAL_FEATURE_GROUP",
                        },
                    )
                elif artifact == Links.Type.FEATURE_VIEW:
                    return Links.__parse_feature_views(
                        links_json["downstream"], {"FEATURE_VIEW"}
                    )
                else:
                    return Links()


class ProvenanceEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Links):
            return {
                "accessible": obj.accessible,
                "inaccessible": obj.inaccessible,
                "deleted": obj.deleted,
                "faulty": obj.faulty,
            }
        elif isinstance(
            obj,
            (
                feature_group.FeatureGroup,
                feature_group.ExternalFeatureGroup,
                feature_view.FeatureView,
                training_dataset.TrainingDataset,
                Artifact,
            ),
        ):
            return {
                "feature_store_name": obj.feature_store_name,
                "name": obj.name,
                "version": obj.version,
            }
        elif isinstance(
            obj,
            (storage_connector.StorageConnector),
        ):
            return {
                "name": obj.name,
            }
        else:
            import importlib.util

            if importlib.util.find_spec("hsml"):
                from hsml import model
                from hsml.core import explicit_provenance as hsml_explicit_provenance

                if isinstance(
                    obj,
                    (
                        model.Model,
                        hsml_explicit_provenance.Artifact,
                    ),
                ):
                    return {
                        "model_registry_id": obj.model_registry_id,
                        "name": obj.name,
                        "version": obj.version,
                    }
            return json.JSONEncoder.default(self, obj)
